from ._download import DownloadCommand
from ._migrate import (
    MigrationCanvasCommand,
    MigrationPrepareCommand,
)
from ._profile import ProfileAssetCentricCommand, ProfileAssetCommand, ProfileRawCommand, ProfileTransformationCommand
from ._purge import PurgeCommand
from ._upload import UploadCommand
from .auth import AuthCommand
from .build_cmd import BuildCommand
from .clean import CleanCommand
from .collect import CollectCommand
from .deploy import DeployCommand
from .dump_resource import DumpResourceCommand
from .featureflag import FeatureFlagCommand
from .init import InitCommand
from .modules import ModulesCommand
from .pull import PullCommand
from .repo import RepoCommand
from .resources import ResourcesCommand
from .run import RunFunctionCommand, RunTransformationCommand, RunWorkflowCommand

__all__ = [
    "AuthCommand",
    "BuildCommand",
    "CleanCommand",
    "CollectCommand",
    "DeployCommand",
    "DownloadCommand",
    "DumpResourceCommand",
    "FeatureFlagCommand",
    "InitCommand",
    "MigrationCanvasCommand",
    "MigrationPrepareCommand",
    "ModulesCommand",
    "ProfileAssetCentricCommand",
    "ProfileAssetCommand",
    "ProfileRawCommand",
    "ProfileTransformationCommand",
    "PullCommand",
    "PurgeCommand",
    "RepoCommand",
    "ResourcesCommand",
    "RunFunctionCommand",
    "RunTransformationCommand",
    "RunWorkflowCommand",
    "UploadCommand",
]



from rich.console import Console

from cognite_toolkit._cdf_tk.client.data_classes.api_classes import PagedResponse
from cognite_toolkit._cdf_tk.client.data_classes.three_d import ThreeDModelResponse
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, ParamRequest
from cognite_toolkit._cdf_tk.utils.useful_types import PrimitiveType


class ThreeDModelAPI:
    ENDPOINT = "/3d/models"

    def __init__(self, http_client: HTTPClient, console: Console) -> None:
        self._http_client = http_client
        self._console = console
        self._config = http_client.config

    def iterate(
        self,
        published: bool | None = None,
        include_revision_info: bool = False,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[ThreeDModelResponse]:
        if not (0 < limit <= 1000):
            raise ValueError("Limit must be between 1 and 1000.")
        parameters: dict[str, PrimitiveType] = {
            # There is a bug in the API. The parameter includeRevisionInfo is expected to be lower case and not
            # camel case as documented. You get error message: Unrecognized query parameter includeRevisionInfo,
            # did you mean includerevisioninfo?
            "includerevisioninfo": include_revision_info,
            "limit": limit,
        }
        if published is not None:
            parameters["published"] = published
        if cursor is not None:
            parameters["cursor"] = cursor
        responses = self._http_client.request_with_retries(
            ParamRequest(
                endpoint_url=self._config.create_api_url(self.ENDPOINT),
                method="GET",
                parameters=parameters,
            )
        )
        responses.raise_for_status()
        return PagedResponse[ThreeDModelResponse].model_validate(responses.get_first_body())


class ThreeDAPI:
    def __init__(self, http_client: HTTPClient, console: Console) -> None:
        self.models = ThreeDModelAPI(http_client, console)



from pathlib import Path

import pytest
import typer
from _pytest.monkeypatch import MonkeyPatch

from cognite_toolkit._cdf_tk.commands import ResourcesCommand
from cognite_toolkit._cdf_tk.constants import MODULES
from cognite_toolkit._cdf_tk.cruds import SpaceCRUD
from tests.test_unit.utils import MockQuestionary


class TestResourcesCreateCommand:
    def test_create_success(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        """Test create method with multiple kinds."""
        cmd = ResourcesCommand(print_warning=False, skip_tracking=True, silent=True)
        organization_dir = tmp_path / "my_org"
        modules_dir = organization_dir / MODULES / "test_module"
        modules_dir.mkdir(parents=True, exist_ok=True)
        (modules_dir / "data_modeling").mkdir(parents=True, exist_ok=True)

        cmd.create(
            organization_dir=organization_dir,
            module_name="test_module",
            kind=["Space", "Container"],
            prefix="my",
            verbose=False,
        )
        assert (modules_dir / "data_modeling" / "my.Space.yaml").exists()
        assert (modules_dir / "data_modeling" / "containers" / "my.Container.yaml").exists()

    def test_create_fuzzy_match_fails_gracefully(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        """Test create method with fuzzy matching for kind suggests and exits."""
        cmd = ResourcesCommand(print_warning=False, skip_tracking=True, silent=True)
        organization_dir = tmp_path / "my_org"
        modules_dir = organization_dir / MODULES / "test_module"
        modules_dir.mkdir(parents=True, exist_ok=True)
        (modules_dir / "data_modeling").mkdir(parents=True, exist_ok=True)

        with pytest.raises(typer.Exit):
            cmd.create(
                organization_dir=organization_dir,
                module_name="test_module",
                kind=["Spce"],
                verbose=False,
            )

    def test_create_unknown_kind_no_match(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        """Test create method with unknown kind and no close match."""
        cmd = ResourcesCommand(print_warning=False, skip_tracking=True, silent=True)
        organization_dir = tmp_path / "my_org"
        modules_dir = organization_dir / MODULES / "test_module"
        modules_dir.mkdir(parents=True, exist_ok=True)
        (modules_dir / "data_modeling").mkdir(parents=True, exist_ok=True)

        with pytest.raises(typer.Exit):
            cmd.create(
                organization_dir=organization_dir,
                module_name="test_module",
                kind=["RocketShip"],
                verbose=False,
            )

    def test_create_module_not_found_create_new(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        """Test create method when module doesn't exist and user creates new."""
        cmd = ResourcesCommand(print_warning=False, skip_tracking=True, silent=True)
        organization_dir = tmp_path / "my_org"

        with MockQuestionary(ResourcesCommand.__module__, monkeypatch, [True]):
            cmd.create(
                organization_dir=organization_dir,
                module_name="new_module",
                kind=["Space"],
                verbose=False,
            )

        modules_dir = organization_dir / MODULES / "new_module"
        assert (modules_dir / "data_modeling" / "my_Space.Space.yaml").exists()

    def test_create_module_not_found_abort(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        """Test create method when module doesn't exist and user aborts."""
        cmd = ResourcesCommand(print_warning=False, skip_tracking=True, silent=True)
        organization_dir = tmp_path / "my_org"

        with pytest.raises(typer.Exit), MockQuestionary(ResourcesCommand.__module__, monkeypatch, [False]):
            cmd.create(
                organization_dir=organization_dir,
                module_name="non_existent",
                kind=["Space"],
                verbose=False,
            )

    def test_create_interactive_module_selection(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        """Test interactive module selection."""
        cmd = ResourcesCommand(print_warning=False, skip_tracking=True, silent=True)
        organization_dir = tmp_path / "my_org"
        modules_dir = organization_dir / MODULES / "existing_module"
        modules_dir.mkdir(parents=True, exist_ok=True)
        (modules_dir / "data_modeling").mkdir(parents=True, exist_ok=True)

        with MockQuestionary(ResourcesCommand.__module__, monkeypatch, [modules_dir]):
            cmd.create(
                organization_dir=organization_dir,
                module_name=None,
                kind=["Space"],
                verbose=False,
            )
        assert (modules_dir / "data_modeling" / "my_Space.Space.yaml").exists()

    def test_create_interactive_new_module(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        """Test interactive selection -> Create new module."""
        cmd = ResourcesCommand(print_warning=False, skip_tracking=True, silent=True)
        organization_dir = tmp_path / "my_org"
        organization_dir.mkdir(parents=True, exist_ok=True)

        with MockQuestionary(ResourcesCommand.__module__, monkeypatch, ["NEW", "created_module"]):
            cmd.create(
                organization_dir=organization_dir,
                module_name=None,
                kind=["Space"],
                verbose=False,
            )

        new_module_path = organization_dir / MODULES / "created_module"
        assert (new_module_path / "data_modeling" / "my_Space.Space.yaml").exists()

    def test_create_interactive_kind_selection(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        """Test interactive kind selection when kind is not provided."""
        cmd = ResourcesCommand(print_warning=False, skip_tracking=True, silent=True)
        organization_dir = tmp_path / "my_org"
        modules_dir = organization_dir / MODULES / "test_module"
        modules_dir.mkdir(parents=True, exist_ok=True)
        (modules_dir / "data_modeling").mkdir(parents=True, exist_ok=True)

        def return_space_crud(*args):
            return SpaceCRUD

        with MockQuestionary(ResourcesCommand.__module__, monkeypatch, [return_space_crud]):
            cmd.create(
                organization_dir=organization_dir,
                module_name="test_module",
                kind=None,
                verbose=False,
            )
        assert (modules_dir / "data_modeling" / "my_Space.Space.yaml").exists()

    def test_create_interactive_kind_selection_abort(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        """Test interactive kind selection when kind is not provided and user aborts."""
        cmd = ResourcesCommand(print_warning=False, skip_tracking=True, silent=True)
        organization_dir = tmp_path / "my_org"
        modules_dir = organization_dir / MODULES / "test_module"
        modules_dir.mkdir(parents=True, exist_ok=True)
        (modules_dir / "data_modeling").mkdir(parents=True, exist_ok=True)

        with pytest.raises(typer.Exit), MockQuestionary(ResourcesCommand.__module__, monkeypatch, [None]):
            cmd.create(
                organization_dir=organization_dir,
                module_name="test_module",
                kind=None,
                verbose=False,
            )



from cognite.client import CogniteClient


def handle(data: dict, client: CogniteClient, secrets: dict, function_call_info: dict) -> dict:
    print("Function 'fn_data_modeling_only_function' called with data:", data)
    return {
        "secrets": mask_secrets(secrets),
        "functionInfo": function_call_info,
    }


def mask_secrets(secrets: dict) -> dict:
    return {k: "***" for k in secrets}



import streamlit as st
from cognite.client import CogniteClient

st.title("An example app in CDF")
client = CogniteClient()


@st.cache_data
def get_assets():
    assets = client.assets.list(limit=1000).to_pandas()
    assets = assets.fillna(0)
    return assets


st.write(get_assets())



from typing import Literal

from .base import BaseModelObject, RequestResource, ResponseResource


class NodeReference(BaseModelObject):
    space: str
    external_id: str


class RevisionStatus(BaseModelObject):
    status: Literal["Queued", "Processing", "Done", "Failed"] | None = None
    revision_id: int | None = None
    created_time: int | None = None
    revision_count: int | None = None
    types: list[str] | None = None


class ThreeDModelRequest(RequestResource):
    name: str


class ThreeDModelClassicRequest(ThreeDModelRequest):
    data_set_id: int | None = None
    metadata: dict[str, str] | None = None


class ThreeDModelDMSRequest(ThreeDModelRequest):
    space: str
    type: Literal["CAD", "PointCloud", "Image360"]
    thumbnail_reference: NodeReference | None = None


class ThreeDModelResponse(ResponseResource[ThreeDModelRequest]):
    name: str
    id: int
    created_time: int
    data_set_id: int | None = None
    metadata: dict[str, str] | None = None
    space: str | None = None
    last_revision_info: RevisionStatus | None = None

    def as_request_resource(self) -> ThreeDModelRequest:
        if self.space is None:
            return ThreeDModelClassicRequest._load(self.dump())
        else:
            return ThreeDModelDMSRequest._load(self.dump())



from cognite.client import CogniteClient
from cognite.client.data_classes.capabilities import FunctionsAcl

# You can import from common.tool and get a CDFClientTool instance
# that can be used to run the function locally and verify capabilities.
from common.tool import CDFClientTool


def handle(data: dict, client: CogniteClient, secrets: dict, function_call_info: dict) -> dict:
    tool = CDFClientTool(client=client)
    # This will fail unless the function has the specified capabilities.
    tool.verify_capabilities(
        [
            FunctionsAcl([FunctionsAcl.Action.Read, FunctionsAcl.Action.Write], FunctionsAcl.Scope.All()),
        ]
    )
    print(tool)
    return {
        "data": data,
        "secrets": mask_secrets(secrets),
        "functionInfo": function_call_info,
    }


def mask_secrets(secrets: dict) -> dict:
    return {k: "***" for k in secrets}



from cognite.client import CogniteClient
from other_module import to_camel, to_pascal, to_snake


def handle(data: dict, client: CogniteClient, function_call_info: dict) -> dict:
    """Convert string to different cases."""
    output: dict[str, str] = {}
    for key, value in data.items():
        if isinstance(value, str):
            output[f"{key}_camel"] = to_camel(value)
            output[f"{key}_snake"] = to_snake(value)
            output[f"{key}_pascal"] = to_pascal(value)
    return output



import pytest

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.api_classes import PagedResponse


@pytest.fixture()
def two_3d_models(toolkit_client: ToolkitClient) -> None:
    client = toolkit_client
    models = client.three_d.models.list(limit=2)
    if len(models) == 0:
        client.three_d.models.create(
            name="integration_test_3d_model",
            data_set_id=None,
            metadata={"source": "integration_test"},
        )
    if len(models) == 1:
        client.three_d.models.create(
            name="integration_test_3d_model_2",
            data_set_id=None,
            metadata={"source": "integration_test"},
        )


class Test3DAPI:
    @pytest.mark.usefixtures("two_3d_models")
    def test_iterate(self, toolkit_client: ToolkitClient) -> None:
        client = toolkit_client
        response = client.tool.three_d.models.iterate(limit=1, include_revision_info=True)
        response2 = client.tool.three_d.models.iterate(limit=1, include_revision_info=True, cursor=response.next_cursor)
        assert isinstance(response, PagedResponse)
        assert isinstance(response2, PagedResponse)
        assert response != response2

        items = response.items + response2.items
        assert 0 < len(items) <= 2
        missing_revision_info = [item for item in items if item.last_revision_info is None]
        assert len(missing_revision_info) == 0



from __future__ import annotations

from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.builders import get_resource_crud
from cognite_toolkit._cdf_tk.cruds import (
    RESOURCE_CRUD_LIST,
    GroupAllScopedCRUD,
    GroupCRUD,
    GroupResourceScopedCRUD,
    ResourceCRUD,
)
from cognite_toolkit._cdf_tk.tk_warnings import ToolkitNotSupportedWarning, ToolkitWarning
from cognite_toolkit._cdf_tk.tk_warnings.fileread import UnknownResourceTypeWarning


class TestGetCRUD:
    @pytest.mark.parametrize(
        "source_path, resource_folder, expected_loader_cls",
        [
            pytest.param(
                Path(f"some_path/{crud_cls.folder_name}/my.{crud_cls.kind}.yaml"),
                crud_cls.folder_name,
                {
                    GroupResourceScopedCRUD: GroupCRUD,
                    GroupAllScopedCRUD: GroupCRUD,
                }.get(crud_cls, crud_cls),
                id=crud_cls.__name__,
            )
            for crud_cls in RESOURCE_CRUD_LIST
        ],
    )
    def test_get_crud_no_warning(
        self, source_path: Path, resource_folder: str, expected_loader_cls: type[ResourceCRUD]
    ) -> None:
        crud_cls, warning = get_resource_crud(source_path, resource_folder)

        assert warning is None
        assert crud_cls is expected_loader_cls

    @pytest.mark.parametrize(
        "source_path, resource_folder, expected_warning_cls",
        [
            pytest.param(
                Path(f"some_path/unknown_folder/my.{GroupCRUD.kind}.yaml"),
                "unknown_folder",
                ToolkitNotSupportedWarning,
                id="Unknown folder, known kind",
            ),
            pytest.param(
                Path("some_path/group/my.UnknownKind.yaml"),
                GroupCRUD.folder_name,
                UnknownResourceTypeWarning,
                id="Known folder, unknown kind",
            ),
        ],
    )
    def test_get_crud_warning(
        self, source_path: Path, resource_folder: str, expected_warning_cls: type[ToolkitWarning]
    ) -> None:
        crud_cls, warning = get_resource_crud(source_path, resource_folder)

        assert crud_cls is None
        assert type(warning) is expected_warning_cls



from collections import Counter
from collections.abc import Iterable
from pathlib import Path
from typing import Any, get_args

import pytest

from cognite_toolkit._cdf_tk.commands._migrate.selectors import AssetCentricMigrationSelector
from cognite_toolkit._cdf_tk.storageio import (
    AssetIO,
    CanvasIO,
    ChartIO,
    DatapointsIO,
    EventIO,
    FileContentIO,
    InstanceIO,
    RawIO,
    StorageIO,
    TimeSeriesIO,
    get_upload_io,
)
from cognite_toolkit._cdf_tk.storageio.selectors import (
    AllChartsSelector,
    AssetCentricFileSelector,
    AssetSubtreeSelector,
    CanvasExternalIdSelector,
    ChartExternalIdSelector,
    ChartOwnerSelector,
    DataPointsFileSelector,
    DataSelector,
    DataSetSelector,
    FileDataModelingTemplateSelector,
    FileMetadataTemplateSelector,
    InstanceFileSelector,
    InstanceSpaceSelector,
    InstanceViewSelector,
    RawTableSelector,
    Selector,
    SelectorAdapter,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses
from cognite_toolkit._cdf_tk.utils.file import read_yaml_file


def example_selector_data() -> Iterable[tuple]:
    yield pytest.param(
        {"type": "rawTable", "table": {"dbName": "my_db", "tableName": "my_table"}},
        RawTableSelector,
        RawIO,
        RawIO.KIND,
        id="RawTableSelector",
    )
    yield pytest.param(
        {
            "type": "instanceView",
            "view": {"space": "my_space", "externalId": "my_view", "version": "v1"},
            "instanceType": "node",
            "instanceSpaces": ["space1", "space2"],
        },
        InstanceViewSelector,
        InstanceIO,
        InstanceIO.KIND,
        id="InstanceViewSelector",
    )
    yield pytest.param(
        {"type": "instanceFile", "datafile": "path/to/file.csv", "validateInstance": True},
        InstanceFileSelector,
        InstanceIO,
        InstanceIO.KIND,
        id="InstanceFileSelector",
    )
    yield pytest.param(
        {"type": "dataSet", "dataSetExternalId": "my_data_set", "kind": "Assets"},
        DataSetSelector,
        AssetIO,
        AssetIO.KIND,
        id="DataSetSelector Assets",
    )
    yield pytest.param(
        {"type": "assetSubtree", "hierarchy": "root/child", "kind": "Assets"},
        AssetSubtreeSelector,
        AssetIO,
        AssetIO.KIND,
        id="AssetSubtreeSelector Assets",
    )
    yield pytest.param(
        {"type": "dataSet", "dataSetExternalId": "my_data_set", "kind": "TimeSeries"},
        DataSetSelector,
        TimeSeriesIO,
        TimeSeriesIO.KIND,
        id="DataSetSelector TimeSeries",
    )
    yield pytest.param(
        {"type": "assetSubtree", "hierarchy": "root/child", "kind": "TimeSeries"},
        AssetSubtreeSelector,
        TimeSeriesIO,
        TimeSeriesIO.KIND,
        id="AssetSubtreeSelector TimeSeries",
    )
    yield pytest.param(
        {"type": "dataSet", "dataSetExternalId": "my_data_set", "kind": "Events"},
        DataSetSelector,
        EventIO,
        EventIO.KIND,
        id="DataSetSelector Events",
    )
    yield pytest.param(
        {"type": "assetSubtree", "hierarchy": "root/child", "kind": "Events"},
        AssetSubtreeSelector,
        EventIO,
        EventIO.KIND,
        id="AssetSubtreeSelector Events",
    )
    yield pytest.param(
        {"type": "chartOwner", "ownerId": "doctrino"},
        ChartOwnerSelector,
        ChartIO,
        ChartIO.KIND,
        id="ChartOwnerSelector",
    )
    yield pytest.param(
        {"type": "allCharts"},
        AllChartsSelector,
        ChartIO,
        ChartIO.KIND,
        id="AllChartSelector",
    )
    yield pytest.param(
        {"type": "assetFile", "datafile": "path/to/file.csv", "kind": "Assets"},
        AssetCentricFileSelector,
        AssetIO,
        AssetIO.KIND,
        id="AssetCentricFileSelector",
    )
    yield pytest.param(
        {
            "type": "instanceSpace",
            "instanceSpace": "myLocation",
            "instanceType": "node",
            "viewId": {
                "space": "my_space",
                "externalId": "MyView",  # Version is optional
            },
        },
        InstanceSpaceSelector,
        InstanceIO,
        InstanceIO.KIND,
        id="InstanceSpaceSelector",
    )
    yield pytest.param(
        {
            "type": "datapointsFile",
            "timestamp_column": "timestamp",
            "columns": [
                {
                    "column_type": "instance",
                    "column": "ts1",
                    "dtype": "numeric",
                    "space": "space1",
                    "external_id": "extId1",
                },
                {
                    "column_type": "externalId",
                    "column": "ts2",
                    "dtype": "string",
                    "external_id": "extId2",
                },
            ],
        },
        DataPointsFileSelector,
        DatapointsIO,
        DatapointsIO.KIND,
        id="DataPointsFileSelector",
    )
    yield pytest.param(
        {
            "type": "chartExternalId",
            "kind": "Charts",
            "externalIds": ["5400ab34-37f1-470d-b965-d81404ac92d8"],
        },
        ChartExternalIdSelector,
        ChartIO,
        ChartIO.KIND,
        id="ChartExternalIdSelector",
    )
    yield pytest.param(
        {
            "type": "canvasExternalId",
            "kind": "IndustrialCanvas",
            "externalIds": ["c1a2b3d4-e5f6-4789-90ab-cdef12345678"],
        },
        CanvasExternalIdSelector,
        CanvasIO,
        CanvasIO.KIND,
        id="CanvasExternalIdSelector",
    )
    yield pytest.param(
        {
            "type": "fileMetadataTemplate",
            "kind": "FileContent",
            "file_directory": "path/to/files",
            "template": {
                "name": "$FILENAME",
                "external_id": "file_$FILENAME",
                "source": "Uploaded via Toolkit",
            },
        },
        FileMetadataTemplateSelector,
        FileContentIO,
        FileContentIO.KIND,
        id="FileMetadataTemplateSelector",
    )
    yield pytest.param(
        {
            "type": "fileDataModelingTemplate",
            "kind": "FileContent",
            "file_directory": "path/to/files",
            "viewId": {
                "space": "cdf_cdm",
                "externalId": "CogniteFile",
                "version": "v1",
            },
            "template": {
                "space": "my_space",
                "external_id": "data_modeling_$FILENAME",
            },
        },
        FileDataModelingTemplateSelector,
        FileContentIO,
        FileContentIO.KIND,
        id="FileDataModelingTemplateSelector",
    )


@pytest.fixture(scope="module")
def all_selectors() -> list[type[DataSelector]]:
    return get_concrete_subclasses(DataSelector)


class TestDataSelectors:
    """Test to ensure all data selectors are working as expected."""

    def test_all_selectors_in_union(self, all_selectors: list[type[DataSelector]]) -> None:
        # The migration selectors are not part of the Selector union, they
        # are only used for migration commands.
        migration_selectors = get_concrete_subclasses(AssetCentricMigrationSelector)
        all_union_selectors = get_args(Selector.__args__[0])
        missing = set(all_selectors) - set(all_union_selectors) - set(migration_selectors)
        assert not missing, (
            f"The following DataSelector subclasses are "
            f"missing from the Selector union: {humanize_collection([cls.__name__ for cls in missing])}"
        )

    def test_all_types_are_unique(self, all_selectors: list[type[DataSelector]]) -> None:
        types = Counter(cls.model_fields["type"].default for cls in all_selectors)
        duplicates = [t for t, count in types.items() if count > 1]
        assert not duplicates, f"The following DataSelector types are not unique: {humanize_collection(duplicates)}"

    def test_example_data_is_complete(self) -> None:
        migration_selectors = get_concrete_subclasses(AssetCentricMigrationSelector)
        # Migration selectors are not part of the Selector union, and are not
        # required to have example data here.
        all_selectors = [cls for cls in get_concrete_subclasses(DataSelector) if cls not in migration_selectors]
        example_types = {p.values[0]["type"] for p in example_selector_data()}
        all_types = {cls.model_fields["type"].default for cls in all_selectors}
        missing = all_types - example_types
        assert not missing, f"The following DataSelector types are missing example data: {humanize_collection(missing)}"

    @pytest.mark.parametrize("data,expected_selector,expected_io,kind", list(example_selector_data()))
    def test_selector_instance(
        self,
        data: dict[str, Any],
        expected_selector: type[DataSelector],
        expected_io: type[StorageIO],
        kind: str,
        tmp_path: Path,
    ) -> None:
        instance = SelectorAdapter.validate_python(data)

        # Assert correct type
        assert isinstance(instance, expected_selector), (
            f"Expected {expected_selector.__name__}, got {type(instance).__name__}"
        )
        # Assert __str__ is implemented
        assert str(instance), f"__str__ not implemented for {type(instance).__name__}"

        # Assert group is implemented
        assert instance.group, f"group property not implemented for {type(instance).__name__}"

        # Assert correct IO type
        assert get_upload_io(instance) is expected_io, (
            f"Expected {expected_io.__name__} for selector {type(instance).__name__}, got {get_upload_io(instance).__name__}"
        )

        # Assert selector is hashable
        assert isinstance(hash(instance), int), f"{type(instance).__name__} is not hashable"

        # Assert serialization/deserialization
        filepath = instance.dump_to_file(tmp_path)
        assert filepath.exists(), f"dump_to_file did not create file for {type(instance).__name__}"

        data = read_yaml_file(filepath)
        loaded = SelectorAdapter.validate_python(data)
        assert loaded.model_dump() == instance.model_dump()
        assert type(loaded) is type(instance)
